==> zoned-0.1.0.tgz: zoned-0.1.0/src/bdb.c <==
/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdint.h>		/* uint32_t */

#include <signal.h>		/* sig_atomic_t */

#include <sys/time.h>		/* struct timeval */
#include <sys/types.h>		/* u_char */

#include <event.h>

#include <db.h>

#include <evnet/thread.h>

#include "options.h"
#include "log.h"
#include "zmbuf.h"
#include "znode.h"
#include "locate.h"
#include "edit.h"
#include "bdb.h"


struct bdb DNSDB = {
	.master	= { .envid = DB_EID_INVALID },
}; /* DNSDB */


static void bdb_master_set(typeof(&DNSDB) bdb, int envid) {
	struct znode *z;

	if (envid <= 0) {
		notice("no master");

		bdb->master.envid	= DB_EID_INVALID;

		return /* void */;
	}

	if (0 == (z = znode_find_byenvid(envid))) {
		warn("newly elected master disappeared");

		return /* void */;
	}

	notice("elected master (%s)", (z == zself)? "self" : znode_label(z));

	bdb->master.envid	= envid;

	return /* void */;
} /* bdb_master_set() */


static void *bdb_dbenv_elect_run(void *cval, void *arg, unsigned timeout) {
	typeof(&DNSDB) bdb	= arg;
	int nsites		= locate_howmany();
	int dberr, masterid;

	if (0 != (dberr = bdb->db.env->rep_elect(bdb->db.env, nsites, 0, &masterid, 0))) {
		notice("election failure: %s", db_strerror(dberr));

		masterid	= DB_EID_INVALID;
	}

	return (void *)(intptr_t)masterid;
} /* bdb_dbenv_select_run() */


static void bdb_dbenv_elect_stop(void *rval, enum thread_errno err, void *arg) {
	typeof(&DNSDB) bdb	= arg;
	int masterid		= (intptr_t)rval;

	notice("election complete");

	bdb_master_set(bdb, masterid);

	bdb_openall(bdb, 0);

	edit_wake();

	return /* void */;
} /* bdb_dbenv_elect_stop() */


static void bdb_dbenv_elect_start(typeof(&DNSDB) bdb) {
	notice("starting election");

	thread_run(&bdb_dbenv_elect_run, 0, THREAD_PRI_NORM, &bdb_dbenv_elect_stop, bdb, 30);
} /* bdb_dbenv_elect_start() */


void zmbuf_recv(struct zmbuf *mb, int envid) {
	typeof(&DNSDB) bdb	= &DNSDB;
	DBT control, rec;
	DB_LSN rlsnp;
	int renvid, dberr;

	control.data	= mb->msg->data.control.data.buf;
	control.size	= mb->msg->data.control.data.size;

	rec.data	= mb->msg->data.rec.data.buf;
	rec.size	= mb->msg->data.rec.data.size;

	renvid		= envid;

	dberr		= bdb->db.env->rep_process_message(bdb->db.env, &control, &rec, &renvid, &rlsnp);

	zmbuf_free(mb), mb = 0;

	switch (dberr) {
	case DB_REP_DUPMASTER:
		debug0("duplicate master");

		debug0("issuing replication start");

		if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, DB_REP_CLIENT)))
			panic("dbenv_rep_start: %s", db_strerror(dberr));

		/* FALL THROUGH */
	case DB_REP_HOLDELECTION:
		bdb_dbenv_elect_start(bdb);

		break;
	case DB_REP_IGNORE:
		debug0("ignoring message");

		break;	
	case DB_REP_ISPERM:
		debug0("got permanent message");

		break;
	case DB_REP_JOIN_FAILURE:
		debug0("join failure");

		break;
	case DB_REP_NEWMASTER:
		debug0("new master: %d", renvid);

		bdb->master.envid	= renvid;

		if (renvid == BDB_SELF_EID) {
			notice("elected master");

			if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, DB_REP_MASTER)))
				panic("dbenv_rep_start: %s", db_strerror(dberr));
		}

		bdb_openall(bdb, 0);

		break;
	case DB_REP_NEWSITE:
		debug0("encountered new host");

		break;
	case DB_REP_NOTPERM:
		debug0("could not commit permanent message");

		break;
	case 0:
		debug0("message processed");

		break;
	default:
		error("could not process message: %s", db_strerror(dberr));	
	}

	return /* void */;
} /* zmbuf_recv() */


static int bdb_dbenv_send(DB_ENV *dbenv, const DBT *control, const DBT *rec, const DB_LSN *lsnp, int envid, uint32_t flags) {
	struct zmbuf *mb;
	int rval;

	if (lsnp == 0)
		panic("lsnp == 0");

	if (0 == (mb = zmbuf_data(control, rec, lsnp))) {
		error("unable to allocate message object");

		return -1;
	}

	if (0 != (rval = zmbuf_send(mb, envid, flags))) {
		debug0("discarding message");

		zmbuf_free(mb), mb = 0;
	}

	return rval;
} /* bdb_dbenv_send() */


static void bdb_dbenv_catch(DB_ENV *dbenv, uint32_t which, void *info) {
	typeof(&DNSDB) bdb	= dbenv->app_private;

	switch (which) {
	case DB_EVENT_REP_MASTER:
		debug("am master");

		bdb->master.is_self	= 1;

		break;
	case DB_EVENT_REP_CLIENT:
		debug("am not master");

		bdb->master.is_self	= 0;

		break;
	case DB_EVENT_REP_STARTUPDONE:
		debug("startup done");

		break;
	case DB_EVENT_REP_NEWMASTER:
		debug("new master");

		break;	
	default:
		debug("ignoring event %d", which);
	} /* switch() */

	return /* void */;
} /* bdb_dbenv_catch() */


void bdb_wakeup(enum bdb_event why) {
	typeof (&DNSDB) bdb	= &DNSDB;
	int dberr;
	int role;

	if (bdb->master.envid > 0 && 0 == znode_find_byenvid(bdb->master.envid))
		bdb_master_set(bdb, DB_EID_INVALID);

	switch (why) {
	case BDB_EV_NODE_JOIN:
	case BDB_EV_NODE_GONE:
		if (bdb->master.envid > 0)
			return /* void */;

		break;
	case BDB_EV_BOTHER:
		if (bdb->master.envid > 0)
			return /* void */;
	}

	if (locate_howmany() > 0)
		bdb->db.env->rep_set_nsites(bdb->db.env, locate_howmany());

	debug("issuing replication start");

	if (locate_howmany() == 2 && znode_count_byhost() == 1)
		role	= DB_REP_MASTER;
	else
		role	= DB_REP_CLIENT;

	if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, role)))
		panic("dbenv_rep_start: %s", db_strerror(dberr));

	bdb_dbenv_elect_start(bdb);

	return /* void */;
} /* bdb_wakeup() */


static void bdbenv_error(const DB_ENV *dbenv, const char *errpfx, const char *msg) {
	debug0("%s", msg);

	return /* void */;
} /* bdbenv_error() */


static void bdb_opendb(typeof(&DNSDB) bdb, DB **db, DBTYPE type, const char *name, const char *file, int optflags, int reopen) {
	int dbflags	= DB_THREAD | DB_AUTO_COMMIT;
	int dbcreat	= (bdb->master.envid == BDB_SELF_EID)? DB_CREATE : 0;
	int dberr;

	if (*db != 0) {
		if (!reopen)
			return /* void */;

		if (0 != (dberr = (*db)->close(*db, 0)))
			panic("db_close: %s", db_strerror(dberr));

		*db	= 0;
	}

	if (0 != (dberr = db_create(db, bdb->db.env, 0)))
		panic("db_create: %s", db_strerror(dberr));

	(*db)->set_errcall(*db, &bdbenv_error);

	if (0 != (dberr = (*db)->set_flags(*db, optflags)))
		panic("db_set_flags: %s", db_strerror(dberr));

	if (0 != (dberr = (*db)->open(*db, 0, file, name, type, dbflags|dbcreat, 0666))) {
		if (dberr != ENOENT && 0 != (dberr = (*db)->close(*db, 0)))
			panic("db_open: %s", db_strerror(dberr));

		*db	= 0;
	}

	return /* void */;
} /* bdb_opendb() */


void bdb_openall(typeof(&DNSDB) bdb, int reopen) {
	luaZ_getglobal(opts.L, "db", "file");

	bdb_opendb(bdb, &bdb->db.data, DB_HASH, "dns_data", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);
	bdb_opendb(bdb, &bdb->db.xfr, DB_BTREE, "dns_xfr", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);
	bdb_opendb(bdb, &bdb->db.zone, DB_BTREE, "dns_zone", luaL_optstring(opts.L, -1, "dnsdata.db"), 0, reopen);
	bdb_opendb(bdb, &bdb->db.client, DB_HASH, "dns_client", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);

	lua_pop(opts.L, 1);

	return /* void */;
} /* bdb_openall() */


static void bdb_timer(int fd, short events, void *arg) {
	bdb_wakeup(BDB_EV_BOTHER);

	timeout_add(arg, (&(struct timeval){ 10, 0 }));

	return /* void */;
} /* bdb_timer() */


void bdb_init(void) {
	typeof(&DNSDB) bdb	= &DNSDB;
	int envflags		= DB_CREATE | DB_INIT_MPOOL | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_REP | DB_INIT_TXN | DB_RECOVER | DB_THREAD;
	int dberr;

	if (0 != (dberr = db_env_create(&bdb->db.env, 0)))
		panic("db_env_create: %s", db_strerror(dberr));

	bdb->db.env->set_errcall(bdb->db.env, &bdbenv_error);

	bdb->db.env->app_private	= bdb;

	if (0 != (dberr = bdb->db.env->set_event_notify(bdb->db.env, &bdb_dbenv_catch)))
		panic("dbenv_set_event_notify: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_transport(bdb->db.env, BDB_SELF_EID, &bdb_dbenv_send)))
		panic("dbenv_rep_set_transport: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_config(bdb->db.env, DB_REP_CONF_DELAYCLIENT, 0)))
		panic("dbenv_rep_set_config: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_config(bdb->db.env, DB_REP_CONF_NOAUTOINIT, 0)))
		panic("dbenv_rep_set_config: %s", db_strerror(dberr));

	luaZ_getglobal(opts.L, "db", "env", "dir");

	if (!lua_isstring(opts.L, -1))
		panic("db.env.dir not specified");

	if (0 != (dberr = bdb->db.env->open(bdb->db.env, lua_tostring(opts.L, -1), envflags, 0666)))
		panic("dbenv_open: %s", db_strerror(dberr));

	lua_pop(opts.L, 1);

	bdb_openall(bdb, 0);

	if (0 != (dberr = DNSDB.db.env->rep_start(DNSDB.db.env, 0, DB_REP_CLIENT)))
		panic("dbenv_rep_start: %s", db_strerror(dberr));

	static struct event evb;

	timeout_set(&evb, &bdb_timer, &evb);
	timeout_add(&evb, (&(struct timeval){ 10, 0 }));

	return /* void */;
} /* bdb_init() */

==> zoned-0.1.1.tgz: zoned-0.1.1/src/bdb.c <==
/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdint.h>		/* uint32_t */

#include <signal.h>		/* sig_atomic_t */

#include <sys/time.h>		/* struct timeval */
#include <sys/types.h>		/* u_char */

#include <event.h>

#include <db.h>

#include <evnet/thread.h>

#include "options.h"
#include "log.h"
#include "zmbuf.h"
#include "znode.h"
#include "locate.h"
#include "edit.h"
#include "bdb.h"


struct bdb DNSDB = {
	.master	= { .envid = DB_EID_INVALID },
}; /* DNSDB */


static void bdb_master_set(typeof(&DNSDB) bdb, int envid) {
	struct znode *z;

	if (envid <= 0) {
		notice("no master");

		bdb->master.envid	= DB_EID_INVALID;

		return /* void */;
	}

	if (0 == (z = znode_find_byenvid(envid))) {
		warn("newly elected master disappeared");

		return /* void */;
	}

	notice("elected master (%s)", (z == zself)? "self" : znode_label(z));

	bdb->master.envid	= envid;

	return /* void */;
} /* bdb_master_set() */


static void *bdb_dbenv_elect_run(void *cval, void *arg, unsigned timeout) {
	typeof(&DNSDB) bdb	= arg;
	int nsites		= locate_howmany();
	int dberr, masterid;

	if (0 != (dberr = bdb->db.env->rep_elect(bdb->db.env, nsites, 0, &masterid, 0))) {
		notice("election failure: %s", db_strerror(dberr));

		masterid	= DB_EID_INVALID;
	}

	return (void *)(intptr_t)masterid;
} /* bdb_dbenv_select_run() */


static void bdb_dbenv_elect_stop(void *rval, enum thread_errno err, void *arg) {
	typeof(&DNSDB) bdb	= arg;
	int masterid		= (intptr_t)rval;

	notice("election complete");

	bdb_master_set(bdb, masterid);

	bdb_openall(bdb, 0);

	edit_wake();

	return /* void */;
} /* bdb_dbenv_elect_stop() */


static void bdb_dbenv_elect_start(typeof(&DNSDB) bdb) {
	notice("starting election");

	thread_run(&bdb_dbenv_elect_run, 0, THREAD_PRI_NORM, &bdb_dbenv_elect_stop, bdb, 30);
} /* bdb_dbenv_elect_start() */


void zmbuf_recv(struct zmbuf *mb, int envid) {
	typeof(&DNSDB) bdb	= &DNSDB;
	DBT control, rec;
	DB_LSN rlsnp;
	int renvid, dberr;

	control.data	= mb->msg->data.control.data.buf;
	control.size	= mb->msg->data.control.data.size;

	rec.data	= mb->msg->data.rec.data.buf;
	rec.size	= mb->msg->data.rec.data.size;

	renvid		= envid;

	dberr		= bdb->db.env->rep_process_message(bdb->db.env, &control, &rec, &renvid, &rlsnp);

	zmbuf_free(mb), mb = 0;

	switch (dberr) {
	case DB_REP_DUPMASTER:
		debug0("duplicate master");

		debug0("issuing replication start");

		if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, DB_REP_CLIENT)))
			panic("dbenv_rep_start: %s", db_strerror(dberr));

		/* FALL THROUGH */
	case DB_REP_HOLDELECTION:
		bdb_dbenv_elect_start(bdb);

		break;
	case DB_REP_IGNORE:
		debug0("ignoring message");

		break;	
	case DB_REP_ISPERM:
		debug0("got permanent message");

		break;
	case DB_REP_JOIN_FAILURE:
		debug0("join failure");

		break;
	case DB_REP_NEWMASTER:
		debug0("new master: %d", renvid);

		bdb->master.envid	= renvid;

		if (renvid == BDB_SELF_EID) {
			notice("elected master");

			if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, DB_REP_MASTER)))
				panic("dbenv_rep_start: %s", db_strerror(dberr));
		}

		bdb_openall(bdb, 0);

		break;
	case DB_REP_NEWSITE:
		debug0("encountered new host");

		break;
	case DB_REP_NOTPERM:
		debug0("could not commit permanent message");

		break;
	case 0:
		debug0("message processed");

		break;
	default:
		error("could not process message: %s", db_strerror(dberr));	
	}

	return /* void */;
} /* zmbuf_recv() */


static int bdb_dbenv_send(DB_ENV *dbenv, const DBT *control, const DBT *rec, const DB_LSN *lsnp, int envid, uint32_t flags) {
	struct zmbuf *mb;
	int rval;

	if (lsnp == 0)
		panic("lsnp == 0");

	if (0 == (mb = zmbuf_data(control, rec, lsnp))) {
		error("unable to allocate message object");

		return -1;
	}

	if (0 != (rval = zmbuf_send(mb, envid, flags))) {
		debug0("discarding message");

		zmbuf_free(mb), mb = 0;
	}

	return rval;
} /* bdb_dbenv_send() */


static void bdb_dbenv_catch(DB_ENV *dbenv, uint32_t which, void *info) {
	typeof(&DNSDB) bdb	= dbenv->app_private;

	switch (which) {
	case DB_EVENT_REP_MASTER:
		debug("am master");

		bdb->master.is_self	= 1;

		break;
	case DB_EVENT_REP_CLIENT:
		debug("am not master");

		bdb->master.is_self	= 0;

		break;
	case DB_EVENT_REP_STARTUPDONE:
		debug("startup done");

		break;
	case DB_EVENT_REP_NEWMASTER:
		debug("new master");

		break;	
	default:
		debug("ignoring event %d", which);
	} /* switch() */

	return /* void */;
} /* bdb_dbenv_catch() */


void bdb_wakeup(enum bdb_event why) {
	typeof (&DNSDB) bdb	= &DNSDB;
	int dberr;
	int role;

	if (bdb->master.envid > 0 && 0 == znode_find_byenvid(bdb->master.envid))
		bdb_master_set(bdb, DB_EID_INVALID);

	switch (why) {
	case BDB_EV_NODE_JOIN:
	case BDB_EV_NODE_GONE:
		if (bdb->master.envid > 0)
			return /* void */;

		break;
	case BDB_EV_BOTHER:
		if (bdb->master.envid > 0)
			return /* void */;
	}

	if (locate_howmany() > 0)
		bdb->db.env->rep_set_nsites(bdb->db.env, locate_howmany());

	debug("issuing replication start");

	if (locate_howmany() == 2 && znode_count_byhost() == 1)
		role	= DB_REP_MASTER;
	else
		role	= DB_REP_CLIENT;

	if (0 != (dberr = bdb->db.env->rep_start(bdb->db.env, 0, role)))
		panic("dbenv_rep_start: %s", db_strerror(dberr));

	bdb_dbenv_elect_start(bdb);

	return /* void */;
} /* bdb_wakeup() */


static void bdbenv_error(const DB_ENV *dbenv, const char *errpfx, const char *msg) {
	debug0("%s", msg);

	return /* void */;
} /* bdbenv_error() */


static void bdb_closedb(struct bdb * bdb, DB **db) {
	int dberr;

	if (*db == 0)
		return /* void */;

	if (0 != (dberr = (*db)->close(*db, 0)))
		panic("db_close: %s", db_strerror(dberr));

	*db	= 0;

	return /* void */;
} /* bdb_closedb() */


static void bdb_opendb(struct bdb *bdb, DB **db, DBTYPE type, const char *name, const char *file, int optflags, int reopen) {
	int dbflags	= DB_THREAD | DB_AUTO_COMMIT | DB_MULTIVERSION;
	int dbcreat	= (bdb->master.envid == BDB_SELF_EID)? DB_CREATE : 0;
	int dberr;

	if (*db != 0) {
		if (!reopen)
			return /* void */;

		bdb_closedb(bdb, db);
	}

	if (0 != (dberr = db_create(db, bdb->db.env, 0)))
		panic("db_create: %s", db_strerror(dberr));

	(*db)->set_errcall(*db, &bdbenv_error);

	if (0 != (dberr = (*db)->set_flags(*db, optflags)))
		panic("db_set_flags: %s", db_strerror(dberr));

	if (0 != (dberr = (*db)->open(*db, 0, file, name, type, dbflags|dbcreat, 0666))) {
		if (dberr != ENOENT && 0 != (dberr = (*db)->close(*db, 0)))
			panic("db_open: %s", db_strerror(dberr));

		*db	= 0;
	}

	return /* void */;
} /* bdb_opendb() */


void bdb_closeall(struct bdb *bdb) {
	bdb_closedb(bdb, &bdb->db.data);
	bdb_closedb(bdb, &bdb->db.xfr);
	bdb_closedb(bdb, &bdb->db.zone);
	bdb_closedb(bdb, &bdb->db.client);

	return /* void */;
} /* bdb_closeall() */


void bdb_openall(struct bdb *bdb, int reopen) {
	luaZ_getglobal(opts.L, "db", "file");

	bdb_opendb(bdb, &bdb->db.data, DB_HASH, "dns_data", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);
	bdb_opendb(bdb, &bdb->db.xfr, DB_BTREE, "dns_xfr", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);
	bdb_opendb(bdb, &bdb->db.zone, DB_BTREE, "dns_zone", luaL_optstring(opts.L, -1, "dnsdata.db"), 0, reopen);
	bdb_opendb(bdb, &bdb->db.client, DB_HASH, "dns_client", luaL_optstring(opts.L, -1, "dnsdata.db"), DB_DUP | DB_DUPSORT, reopen);

	lua_pop(opts.L, 1);

	return /* void */;
} /* bdb_openall() */


static void bdb_timer(int fd, short events, void *arg) {
	bdb_wakeup(BDB_EV_BOTHER);

	timeout_add(arg, (&(struct timeval){ 10, 0 }));

	return /* void */;
} /* bdb_timer() */


void bdb_init(void) {
	typeof(&DNSDB) bdb	= &DNSDB;
	int envflags		= DB_CREATE | DB_INIT_MPOOL | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_REP | DB_INIT_TXN | DB_RECOVER | DB_THREAD;
	int dberr;

	if (0 != (dberr = db_env_create(&bdb->db.env, 0)))
		panic("db_env_create: %s", db_strerror(dberr));

	bdb->db.env->set_errcall(bdb->db.env, &bdbenv_error);

	bdb->db.env->app_private	= bdb;

	if (0 != (dberr = bdb->db.env->set_event_notify(bdb->db.env, &bdb_dbenv_catch)))
		panic("dbenv_set_event_notify: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_transport(bdb->db.env, BDB_SELF_EID, &bdb_dbenv_send)))
		panic("dbenv_rep_set_transport: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_config(bdb->db.env, DB_REP_CONF_DELAYCLIENT, 0)))
		panic("dbenv_rep_set_config: %s", db_strerror(dberr));

	if (0 != (dberr = bdb->db.env->rep_set_config(bdb->db.env, DB_REP_CONF_NOAUTOINIT, 0)))
		panic("dbenv_rep_set_config: %s", db_strerror(dberr));

	luaZ_getglobal(opts.L, "db", "env", "dir");

	if (!lua_isstring(opts.L, -1))
		panic("db.env.dir not specified");

	if (0 != (dberr = bdb->db.env->open(bdb->db.env, lua_tostring(opts.L, -1), envflags, 0666)))
		panic("dbenv_open: %s", db_strerror(dberr));

	lua_pop(opts.L, 1);

	bdb_openall(bdb, 0);

	if (0 != (dberr = DNSDB.db.env->rep_start(DNSDB.db.env, 0, DB_REP_CLIENT)))
		panic("dbenv_rep_start: %s", db_strerror(dberr));

	static struct event evb;

	timeout_set(&evb, &bdb_timer, &evb);
	timeout_add(&evb, (&(struct timeval){ 10, 0 }));

	return /* void */;
} /* bdb_init() */
